@@ -3,8 +3,9 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2010-2015, University of Amsterdam
+    Copyright (c)  2010-2020, University of Amsterdam
                               VU University Amsterdam
+                              SWI-Prolog Solutions b.v.
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -39,120 +40,210 @@
             sparql_read_xml_result/2,   % +Stream, -Result
             sparql_read_json_result/2   % +Input, -Result
           ]).
-:- use_module(library(http/http_open)).
-:- use_module(library(http/json)).
-:- use_module(library(lists)).
-:- use_module(library(rdf)).
-:- use_module(library(semweb/rdf_turtle)).
-:- use_module(library(option)).
+:- autoload(library(apply), [maplist/3, maplist/4, partition/4]).
+:- autoload(library(gensym), [gensym/2]).
+:- autoload(library(lists), [member/2]).
+:- autoload(library(option), [select_option/3, select_option/4, merge_options/3]).
+:- autoload(library(rdf), [load_rdf/2]).
+:- autoload(library(readutil), [read_stream_to_codes/2]).
+:- autoload(library(sgml), [load_structure/3]).
+:- autoload(library(uri),
+            [ uri_components/2,
+              uri_data/3,
+              uri_authority_components/2,
+              uri_authority_data/3
+            ]).
+:- autoload(library(http/http_open), [http_open/3]).
+:- autoload(library(http/json), [json_read/2]).
+:- autoload(library(semweb/turtle), [rdf_read_turtle/3]).
 
 /** <module> SPARQL client library
 
 This module provides a SPARQL client.  For example:
 
-    ==
-    ?- sparql_query('select * where { ?x rdfs:label "Amsterdam" }', Row,
-                    [ host('dbpedia.org'), path('/sparql/')]).
+```
+?- sparql_query('select * where { ?x rdfs:label "Amsterdam" }', Row,
+                [ host('dbpedia.org'), path('/sparql/')]).
 
-    Row = row('http://www.ontologyportal.org/WordNet#WN30-108949737') ;
-    false.
-    ==
+Row = row('http://www.ontologyportal.org/WordNet#WN30-108949737') ;
+false.
+```
 
-Or, querying a local server using an =ASK= query:
+Or, querying a local server using an ``ASK`` query:
 
-    ==
-    ?- sparql_query('ask { owl:Class rdfs:label "Class" }', Row,
-                    [ host('localhost'), port(3020), path('/sparql/')]).
-    Row = true.
-    ==
+```
+?- sparql_query('ask { owl:Class rdfs:label "Class" }', Row,
+                [ host('localhost'), port(3020), path('/sparql/')]).
+Row = true.
+```
+
+HTTPS servers are supported using the scheme(https) option:
+
+```
+?- sparql_query('select * where { ?x rdfs:label "Amsterdam"@nl }',
+                Row,
+                [ scheme(https),
+                  host('query.wikidata.org'),
+                  path('/sparql')
+                ]).
+```
 */
 
 
 %!  sparql_query(+Query, -Result, +Options) is nondet.
 %
-%   Execute a SPARQL query on an HTTP   SPARQL endpoint. Query is an
-%   atom that denotes  the  query.  Result   is  unified  to  a term
-%   rdf(S,P,O) for =CONSTRUCT= and =DESCRIBE=  queries, row(...) for
-%   =SELECT= queries and  =true=  or   =false=  for  =ASK=  queries.
-%   Options are
+%   Execute a SPARQL query on an HTTP  SPARQL endpoint. Query is an atom
+%   that denotes the query. Result is unified   to a term rdf(S,P,O) for
+%   ``CONSTRUCT`` and ``DESCRIBE``  queries,   row(...)  for  ``SELECT``
+%   queries and `true` or `false` for ``ASK`` queries. Options are
 %
-%       * host(+Host)
-%       * port(+Port)
-%       * path(+Path)
-%       The above three options set the location of the server.
-%       * search(+ListOfParams)
-%       Provide additional query parameters, such as the graph.
-%       * variable_names(-ListOfNames)
-%       Unifies ListOfNames with a list of atoms that describe the
-%       names of the variables in a =SELECT= query.
+%   Variables that are unbound in SPARQL   (e.g., due to SPARQL optional
+%   clauses), are bound in Prolog to the atom `'$null$'`.
 %
-%   Remaining options are passed to   http_open/3.  The defaults for
-%   Host, Port and Path can be   set  using sparql_set_server/1. The
-%   initial default for port is 80 and path is =|/sparql/|=.
+%	- endpoint(+URL)
+%	  May be used as alternative to Scheme, Host, Port and Path
+%	  to specify the endpoint in a single option.
+%       - host(+Host)
+%       - port(+Port)
+%       - path(+Path)
+%       - scheme(+Scheme)
+%         The above four options set the location of the server.
+%       - search(+ListOfParams)
+%         Provide additional query parameters, such as the graph.
+%       - variable_names(-ListOfNames)
+%         Unifies ListOfNames with a list of atoms that describe the
+%         names of the variables in a =SELECT= query.
 %
-%   For example, the ClioPatria  server   understands  the parameter
-%   =entailment=. The code  below  queries   for  all  triples using
+%   Remaining options are passed to http_open/3.  The defaults for Host,
+%   Port and Path can be  set   using  sparql_set_server/1.  The initial
+%   default for port is 80 and path is `/sparql/`.
+%
+%   For  example,  the  ClioPatria  server   understands  the  parameter
+%   `entailment`.  The  code  below  queries    for  all  triples  using
 %   _rdfs_entailment.
 %
-%     ==
-%     ?- sparql_query('select * where { ?s ?p ?o }',
-%                     Row,
-%                     [ search([entailment=rdfs])
-%                     ]).
-%     ==
+%   ```
+%   ?- sparql_query('select * where { ?s ?p ?o }',
+%                   Row,
+%                   [ search([entailment=rdfs])
+%                   ]).
+%   ```
+%
+%   Another useful option is the   `request_header`  which, for example,
+%   may be used to trick force  a   server  to  reply using a particular
+%   document format:
+%
+%   ```
+%   ?- sparql_query(
+%          'select * where { ?s ?p ?o }',
+%           Row,
+%           [ host('integbio.jp'),
+%             path('/rdf/sparql'),
+%             request_header('Accept' =
+%                            'application/sparql-results+xml')
+%           ]).
+%   ```
 
 sparql_query(Query, Row, Options) :-
-    sparql_param(host(Host), Options,  Options1),
-    sparql_param(port(Port), Options1, Options2),
-    sparql_param(path(Path), Options2, Options3),
-    select_option(search(Extra), Options3, Options4, []),
-    select_option(variable_names(VarNames), Options4, Options5, _),
-    sparql_extra_headers(HTTPOptions),
-    http_open([ protocol(http),
+    (   select_option(endpoint(URL), Options, Options5)
+    ->  uri_components(URL, Components),
+        uri_data(scheme, Components, Scheme),
+        uri_data(authority, Components, Auth),
+        uri_data(path, Components, Path),
+        uri_data(search, Components, Extra),
+        ignore(Extra = []),
+        uri_authority_components(Auth, AComp),
+        uri_authority_data(host, AComp, Host),
+        uri_authority_data(port, AComp, Port),
+        (   var(Port)
+        ->  sparql_port(Scheme, Port, _, _)
+        ;   true
+        )
+    ;   sparql_param(scheme(Scheme), Options,  Options1),
+        sparql_port(Scheme, Port,    Options1, Options2),
+        sparql_param(host(Host),     Options2, Options3),
+        sparql_param(path(Path),     Options3, Options4),
+        select_option(search(Extra), Options4, Options5, [])
+    ),
+    select_option(variable_names(VarNames), Options5, Options6, _),
+    partition(is_url_option, Options6, UrlOptions, HTTPOptions),
+    sparql_extra_headers(HTTPOptions0),
+    merge_options(HTTPOptions, HTTPOptions0, HTTPOptions1),
+    http_open([ scheme(Scheme),
                 host(Host),
                 port(Port),
                 path(Path),
                 search([ query = Query
                        | Extra
                        ])
-              | Options5
+              | UrlOptions
               ], In,
-              [ header(content_type, ContentType)
-              | HTTPOptions
+              [ header(content_type, ContentType),
+                status_code(Status)
+              | HTTPOptions1
               ]),
     plain_content_type(ContentType, CleanType),
-    read_reply(CleanType, In, VarNames, Row).
+    read_reply(Status, CleanType, In, VarNames, Row).
+
+url_option(scheme).
+url_option(user).
+url_option(password).
+url_option(host).
+url_option(port).
+url_option(path).
+url_option(query_string).
+url_option(search).
+url_option(cert_verify_hook).
+url_option(certificate_file).
+url_option(key_file).
+url_option(certificate_key_pairs).
+url_option(pem_password_hook).
+url_option(crl).
+url_option(cacert_file).
+url_option(cacerts).
+url_option(cipher_list).
+url_option(ecdh_curve).
+url_option(min_protocol_version).
+url_option(max_protocol_version).
+url_option(disable_ssl_methods).
+url_option(ssl_method).
+
+is_url_option(Name = _Value) :-
+    url_option(Name),
+    !.
+is_url_option(Opt) :-
+    compound(Opt),
+    functor(Opt, Name, 1),
+    url_option(Name).
 
 %!  sparql_extra_headers(-List)
 %
 %   Send extra headers with the request. Note that, although we also
-%   process RDF embedded in HTML, we do  not explicitely ask for it.
+%   process RDF embedded in HTML, we do  not explicitly ask for it.
 %   Doing so causes some   (e.g., http://w3.org/2004/02/skos/core to
 %   reply with the HTML description rather than the RDF).
 
 sparql_extra_headers(
         [ request_header('Accept' = 'application/sparql-results+xml, \c
                                      application/n-triples, \c
-                                     application/x-turtle, \c
-                                     application/turtle, \c
+                                     application/x-turtle; q=0.9, \c
+                                     application/turtle; q=0.9, \c
+                                     text/turtle, \c
                                      application/sparql-results+json, \c
-                                     text/turtle; q=0.9, \c
                                      application/rdf+xml, \c
                                      text/rdf+xml; q=0.8, \c
-                                     */*; q=0.1'),
-          cert_verify_hook(ssl_verify)
+                                     */*; q=0.1')
         ]).
 
-:- public ssl_verify/5.
-
-%!  ssl_verify(+SSL, +ProblemCert, +AllCerts, +FirstCert, +Error)
-%
-%   Currently we accept  all  certificates.
-
-ssl_verify(_SSL,
-           _ProblemCertificate, _AllCertificates, _FirstCertificate,
-           _Error).
+%!  read_reply(+Status, +ContentType, +In, -Close, -Row)
 
+read_reply(200, ContentType, In, Close, Row) :-
+    !,
+    read_reply(ContentType, In, Close, Row).
+read_reply(Status, _ContentType, In, _Close, _Row) :-
+    call_cleanup(read_string(In, _, Reply),
+                 close(In, [force(true)])),
+    throw(error(sparql_error(Status, Reply), _)).
 
 read_reply('application/rdf+xml', In, _, Row) :-
     !,
@@ -225,7 +316,7 @@ varnames(select(VarTerm, _Rows), VarNames) :-
 :- dynamic
     sparql_setting/1.
 
-sparql_setting(port(80)).
+sparql_setting(scheme(http)).
 sparql_setting(path('/sparql/')).
 
 sparql_param(Param, Options0, Options) :-
@@ -238,6 +329,18 @@ sparql_param(Param, Options, Options) :-
     functor(Param, Name, _),
     throw(error(existence_error(option, Name), _)).
 
+sparql_port(_Scheme, Port, Options0, Options) :-
+    select_option(port(Port), Options0, Options),
+    !.
+sparql_port(_Scheme, Port, Options, Options) :-
+    sparql_setting(port(Port)),
+    !.
+sparql_port(http, 80, Options, Options) :-
+    !.
+sparql_port(https, 443, Options, Options) :-
+    !.
+
+
 %!  sparql_set_server(+OptionOrList)
 %
 %   Set sparql server default options.  Provided defaults are:
@@ -329,8 +432,7 @@ term_expansion(T0, T) :-
 
 sparql_read_xml_result(Input, Result) :-
     load_structure(Input, DOM,
-                   [ dialect(xmlns),
-                     space(remove)
+                   [ dialect(xmlns)
                    ]),
     call_cleanup(dom_to_result(DOM, Result),
                  retractall(bnode_map(_,_))).
@@ -362,14 +464,23 @@ variables([element(sparql:variable, Att, [])|T0], [Name|T]) :-
     memberchk(name=Name, Att),
     variables(T0, T).
 variables([element(sparql:link, _, _)|T0], T) :-
+    !,
+    variables(T0, T).
+variables([CDATA|T0], T) :-
+    atomic(CDATA),
     variables(T0, T).
 
 
 rows([], _, []).
 rows([R|T0], Vars, [Row|T]) :-
+    R = element(sparql:result, _, _),
+    !,
     row_values(Vars, R, Values),
     Row =.. [row|Values],
     rows(T0, Vars, T).
+rows([CDATA|T0], Vars, T) :-
+    atomic(CDATA),
+    rows(T0, Vars, T).
 
 row_values([], _, []).
 row_values([Var|VarT], DOM, [Value|ValueT]) :-
@@ -380,8 +491,9 @@ row_values([Var|VarT], DOM, [Value|ValueT]) :-
     ),
     row_values(VarT, DOM, ValueT).
 
-value([element(sparql:literal, Att, Content)], literal(Lit)) :-
+value([element(sparql:literal, Att, Content)|Rest], literal(Lit)) :-
     !,
+    white(Rest),
     lit_value(Content, Value),
     (   memberchk(datatype=Type, Att)
     ->  Lit = type(Type, Value)
@@ -389,13 +501,25 @@ value([element(sparql:literal, Att, Content)], literal(Lit)) :-
     ->  Lit = lang(Lang, Value)
     ;   Lit = Value
     ).
-value([element(sparql:uri, [], [URI])], URI) :- !.
-value([element(sparql:bnode, [], [NodeID])], URI) :-
+value([element(sparql:uri, [], [URI])|Rest], URI) :- !,
+    white(Rest).
+value([element(sparql:bnode, [], [NodeID])|Rest], URI) :-
     !,
+    white(Rest),
     bnode(NodeID, URI).
-value([element(sparql:unbound, [], [])], '$null$').
+value([element(sparql:unbound, [], [])|Rest], '$null$') :-
+    !,
+    white(Rest).
+value([CDATA|Rest], Value) :-
+    atomic(CDATA),
+    value(Rest, Value).
 
 
+white([]).
+white([CDATA|T]) :-
+    atomic(CDATA),
+    white(T).
+
 lit_value([], '').
 lit_value([Value], Value).
 
