@@ -37,6 +37,16 @@
 #include <h/graphics.h>
 #include "include.h"
 
+#ifdef HAVE_X11_EXTENSIONS_XRANDR_H
+#ifdef HAVE_LIBXRANDR
+#define Connection XRRConnection
+#include <X11/extensions/Xrandr.h>
+#undef Connection
+#else
+#undef HAVE_X11_EXTENSIONS_XRANDR_H
+#endif
+#endif
+
 #define X11LastEventTime() ((Time)LastEventTime())
 #define utf8_get_uchar(s, chr) (unsigned char*)utf8_get_char((char *)(s), chr)
 
@@ -146,12 +156,43 @@ ws_resolution_display(DisplayObj d, int *rx, int *ry)
   int screen, wpx, wmm, hpx, hmm;
 
   screen = XDefaultScreen(r->display_xref);
-  wpx = XDisplayWidth(r->display_xref,  screen);
-  hpx = XDisplayHeight(r->display_xref, screen);
-  wmm = XDisplayWidthMM(r->display_xref,  screen);
-  hmm = XDisplayHeightMM(r->display_xref, screen);
-  *rx = (int)((double)wpx*25.4 / (double)wmm + 0.5);
-  *ry = (int)((double)hpx*25.4 / (double)hmm + 0.5);
+
+#ifdef HAVE_X11_EXTENSIONS_XRANDR_H
+  int nsizes = 0;
+  XRRScreenSize *ss = XRRSizes(r->display_xref,  screen, &nsizes);
+
+  if ( nsizes > 0 )
+  { DEBUG(NAME_scale, Cprintf("Getting dimensions from Xrandr extension\n"));
+    wpx = ss[0].width;
+    hpx = ss[0].height;
+    wmm = ss[0].mwidth;
+    hmm = ss[0].mheight;
+  } else
+#endif
+  { DEBUG(NAME_scale, Cprintf("Getting dimensions from Xlib\n"));
+    wpx = XDisplayWidth(r->display_xref,  screen);
+    hpx = XDisplayHeight(r->display_xref, screen);
+    wmm = XDisplayWidthMM(r->display_xref,  screen);
+    hmm = XDisplayHeightMM(r->display_xref, screen);
+  }
+
+  DEBUG(NAME_scale, Cprintf(" --> %dx%d pixels, %dx%dmm\n",
+			    wpx, hpx, wmm, hmm));
+
+  if ( wmm && hmm )
+  { int xdpi = (int)((double)wpx*25.4 / (double)wmm + 0.5);
+    int ydpi = (int)((double)hpx*25.4 / (double)hmm + 0.5);
+
+    double ratio = (double)xdpi/(double)ydpi;
+    if ( ratio > 0.9 && ratio < 1.1 )
+    { *rx = xdpi;
+      *ry = ydpi;
+      succeed;
+    }
+  }
+
+  DEBUG(NAME_scale, Cprintf("Dimensions seem bogus.  Assuming 100x100DPI\n"));
+  *rx = *ry = 100;
 
   succeed;
 }
@@ -742,6 +783,7 @@ ws_get_selection(DisplayObj d, Name which, Name target)
 
   selection_complete = FALSE;
   selection_error = NIL;
+  selection_value = NIL;
   XtGetSelectionValue(r->shell_xref,
 		      nameToSelectionAtom(d, which),
 		      nameToSelectionAtom(d, target),
@@ -749,12 +791,13 @@ ws_get_selection(DisplayObj d, Name which, Name target)
 		      d,
 		      X11LastEventTime());
 
-  while(!selection_complete)
-    dispatchDisplayManager(d->display_manager, DEFAULT, toInt(50));
+  for(int i=0; i<10 && !selection_complete; i++)
+  { dispatchDisplayManager(d->display_manager, DEFAULT, toInt(50));
 
-  if ( notNil(selection_error) )
-  { errorPce(d, NAME_getSelection, which, selection_error);
-    fail;
+    if ( notNil(selection_error) )
+    { errorPce(d, NAME_getSelection, which, selection_error);
+      fail;
+    }
   }
 
   answer(selection_value);
